#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include "asapo/unittests/MockHttpClient.h"
#include "asapo/unittests/MockLogger.h"

#include "../../src/receiver_error.h"
#include "../../src/request.h"
#include "../../src/request_handler/request_handler.h"
#include "../../src/request_handler/request_handler_authorize.h"
#include "asapo/common/networking.h"
#include "../mock_receiver_config.h"
#include "asapo/preprocessor/definitions.h"

#include "../receiver_mocking.h"

#include "../../src/receiver_config.h"


using ::testing::Test;
using ::testing::Return;
using ::testing::ReturnRef;
using ::testing::_;
using ::testing::DoAll;
using ::testing::SetArgReferee;
using ::testing::Gt;
using ::testing::Eq;
using ::testing::Ne;
using ::testing::Mock;
using ::testing::NiceMock;
using ::testing::InSequence;
using ::testing::SetArgPointee;
using ::testing::AllOf;
using ::testing::HasSubstr;

using asapo::MockRequest;
using ::asapo::Error;
using ::asapo::ErrorInterface;
using ::asapo::FileDescriptor;
using ::asapo::SocketDescriptor;
using ::asapo::MockHttpClient;
using asapo::Request;
using asapo::RequestHandlerAuthorize;
using ::asapo::GenericRequestHeader;
using asapo::ReceiverConfig;
using asapo::SetReceiverConfig;
using asapo::HttpCode;

namespace {

TEST(Authorizer, Constructor) {
    RequestHandlerAuthorize handler;
    ASSERT_THAT(dynamic_cast<asapo::HttpClient*>(handler.http_client__.get()), Ne(nullptr));
    ASSERT_THAT(dynamic_cast<const asapo::AbstractLogger*>(handler.log__), Ne(nullptr));
}


class AuthorizerHandlerTests : public Test {
  public:
    RequestHandlerAuthorize handler;
    MockHttpClient mock_http_client;
    std::unique_ptr<MockRequest> mock_request;
    ReceiverConfig config;

    NiceMock<asapo::MockLogger> mock_logger;
    std::string expected_beamtime_id = "beamtime_id";
    std::string expected_data_source = "source";
    std::string expected_beamline = "beamline";
    std::string expected_beamline_path = "/beamline/p01/current";
    std::string expected_core_path = "/gpfs/blabla";
    std::string expected_producer_uri = "producer_uri";
    std::string expected_authorization_server = "authorizer_host";
    std::string expect_request_string;
    std::string expected_source_credentials;
    std::string expected_api_version = "v0.1";

    asapo::SourceType expected_source_type = asapo::SourceType::kProcessed;
    std::string expected_source_type_str = "processed";
    std::string expected_access_type_str = "[\"write\"]";
    void MockRequestData();
    void SetUp() override {
        GenericRequestHeader request_header;
        expected_source_credentials = "processed%" + expected_beamtime_id + "%source%token";
        expect_request_string = std::string("{\"SourceCredentials\":\"") + expected_source_credentials +
                                "\",\"OriginHost\":\"" +
                                expected_producer_uri + "\"}";

        mock_request.reset(new MockRequest{request_header, 1, expected_producer_uri, nullptr});
        handler.http_client__ = std::unique_ptr<asapo::HttpClient> {&mock_http_client};
        handler.log__ = &mock_logger;
        config.authorization_server = expected_authorization_server;
        config.authorization_interval_ms = 0;
        SetReceiverConfig(config, "none");
    }
    void TearDown() override {
        handler.http_client__.release();
    }
    void MockAuthRequest(bool error, HttpCode code = HttpCode::OK, bool opError = false) {
        if (error) {
            EXPECT_CALL(mock_http_client, Post_t(expected_authorization_server + "/authorize", _, expect_request_string, _, _)).
            WillOnce(
                DoAll(SetArgPointee<4>(new asapo::SimpleError("http error")),
                      Return("")
                     ));
            EXPECT_CALL(mock_logger, Error(AllOf(HasSubstr("failure authorizing"),
                                                 HasSubstr("http error"),
                                                 HasSubstr(expected_beamtime_id),
                                                 HasSubstr(expected_producer_uri),
                                                 HasSubstr(expected_authorization_server))));

        } else {
            EXPECT_CALL(mock_http_client, Post_t(expected_authorization_server + "/authorize", _, expect_request_string, _, _)).
            WillOnce(
                DoAll(SetArgPointee<4>(nullptr),
                      SetArgPointee<3>(code),
                      Return("{\"beamtimeId\":\"" + expected_beamtime_id +
                             "\",\"dataSource\":" + "\"" + expected_data_source +
                             "\",\"beamline-path\":" + "\"" + expected_beamline_path +
                             "\",\"corePath\":" + "\"" + expected_core_path +
                             "\",\"source-type\":" + "\"" + expected_source_type_str +
                             "\",\"beamline\":" + "\"" + expected_beamline +
                             "\",\"access-types\":" + expected_access_type_str + "}")
                     ));
            if (code != HttpCode::OK) {
                EXPECT_CALL(mock_logger, Error(AllOf(HasSubstr("failure authorizing"),
                                                     HasSubstr(expected_source_type_str),
                                                     HasSubstr(expected_beamtime_id),
                                                     HasSubstr(expected_data_source),
                                                     HasSubstr(expected_producer_uri),
                                                     HasSubstr(expected_authorization_server))));
            } else if (!opError) {
                EXPECT_CALL(mock_logger, Debug(AllOf(HasSubstr("authorized"),
                                                     HasSubstr(expected_beamtime_id),
                                                     HasSubstr(expected_beamline),
                                                     HasSubstr(expected_source_type_str),
                                                     HasSubstr(expected_data_source),
                                                     HasSubstr(expected_producer_uri))));
            } else {
                EXPECT_CALL(mock_logger, Error(HasSubstr("wrong")));
            }
        }


    }
    Error MockFirstAuthorization(bool error, HttpCode code = HttpCode::OK, bool opError = false) {
        EXPECT_CALL(*mock_request, GetOpCode())
        .WillOnce(Return(asapo::kOpcodeAuthorize))
        ;
        EXPECT_CALL(*mock_request, GetMetaData())
        .WillOnce(ReturnRef(expected_source_credentials))
        ;

        EXPECT_CALL(*mock_request, GetApiVersion())
        .WillOnce(Return(expected_api_version))
        ;


        MockAuthRequest(error, code, opError);
        return handler.ProcessRequest(mock_request.get());
    }
    Error MockRequestAuthorization(bool error, HttpCode code = HttpCode::OK, bool set_request = true) {
        EXPECT_CALL(*mock_request, GetOpCode())
        .WillOnce(Return(asapo::kOpcodeTransferData))
        ;

        if (!error && code == HttpCode::OK && set_request) {
            EXPECT_CALL(*mock_request, SetBeamtimeId(expected_beamtime_id));
            EXPECT_CALL(*mock_request, SetDataSource(expected_data_source));
            EXPECT_CALL(*mock_request, SetOfflinePath(expected_core_path));
            EXPECT_CALL(*mock_request, SetOnlinePath(expected_beamline_path));
            EXPECT_CALL(*mock_request, SetBeamline(expected_beamline));
            EXPECT_CALL(*mock_request, SetSourceType(expected_source_type));
        }

        MockAuthRequest(error, code);
        return handler.ProcessRequest(mock_request.get());
    }

};

TEST_F(AuthorizerHandlerTests, CheckStatisticEntity) {
    auto entity = handler.GetStatisticEntity();
    ASSERT_THAT(entity, Eq(asapo::StatisticEntity::kNetwork));
}

TEST_F(AuthorizerHandlerTests, ErrorNotAuthorizedYet) {
    EXPECT_CALL(*mock_request, GetOpCode())
    .WillOnce(Return(asapo::kOpcodeTransferData))
    ;

    auto err = handler.ProcessRequest(mock_request.get());

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kAuthorizationFailure));
}

TEST_F(AuthorizerHandlerTests, ErrorProcessingAuthorizeRequest) {

    auto err = MockFirstAuthorization(true);

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kInternalServerError));
}


TEST_F(AuthorizerHandlerTests, AuthorizeRequestreturns401) {

    auto err = MockFirstAuthorization(false, HttpCode::Unauthorized);

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kAuthorizationFailure));
}


TEST_F(AuthorizerHandlerTests, AuthorizeOk) {
    auto err = MockFirstAuthorization(false);

    ASSERT_THAT(err, Eq(nullptr));
}


TEST_F(AuthorizerHandlerTests, AuthorizeFailsOnWrongAccessType) {
    expected_access_type_str = "[\"read\"]";
    auto err = MockFirstAuthorization(false, HttpCode::OK, true);

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kAuthorizationFailure));
}

TEST_F(AuthorizerHandlerTests, AuthorizeFailsOnWrongAccessTypeForRaw) {
    expected_access_type_str = "[\"write\"]";
    expected_source_type_str = "raw";
    auto err = MockFirstAuthorization(false, HttpCode::OK, true);
    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kAuthorizationFailure));
}

TEST_F(AuthorizerHandlerTests, AuthorizeFailsOnWrongAccessTypeForProcessed) {
    expected_access_type_str = "[\"writeraw\"]";
    expected_source_type_str = "processed";
    auto err = MockFirstAuthorization(false, HttpCode::OK, true);
    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kAuthorizationFailure));
}

TEST_F(AuthorizerHandlerTests, AuthorizeOkForRaw) {
    expected_access_type_str = "[\"writeraw\"]";
    expected_source_type_str = "raw";
    auto err = MockFirstAuthorization(false, HttpCode::OK, false);
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(AuthorizerHandlerTests, ErrorOnSecondAuthorize) {
    MockFirstAuthorization(false);
    EXPECT_CALL(*mock_request, GetOpCode())
    .WillOnce(Return(asapo::kOpcodeAuthorize));

    EXPECT_CALL(mock_logger, Error(AllOf(HasSubstr("failure authorizing"),
                                         HasSubstr("already authorized"),
                                         HasSubstr(expected_authorization_server))));


    auto err =  handler.ProcessRequest(mock_request.get());

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kAuthorizationFailure));
}

TEST_F(AuthorizerHandlerTests, ErrorOnDataTransferRequestAuthorize) {
    MockFirstAuthorization(false);

    auto err = MockRequestAuthorization(true);

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kInternalServerError));
}


TEST_F(AuthorizerHandlerTests, DataTransferRequestAuthorizeReturns401) {
    MockFirstAuthorization(false);

    auto err = MockRequestAuthorization(false, HttpCode::Unauthorized);

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kReAuthorizationFailure));
}

TEST_F(AuthorizerHandlerTests, DataTransferRequestAuthorizeReturnsSameBeamtimeId) {
    MockFirstAuthorization(false);
    auto err = MockRequestAuthorization(false);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(AuthorizerHandlerTests, RequestAuthorizeReturnsDifferentBeamtimeId) {
    MockFirstAuthorization(false);

    expected_beamtime_id = "different_id";
    auto err = MockRequestAuthorization(false, HttpCode::OK, false);

    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kReAuthorizationFailure));
}


TEST_F(AuthorizerHandlerTests, RequestFromUnsupportedClient) {
    EXPECT_CALL(*mock_request, GetOpCode())
    .WillOnce(Return(asapo::kOpcodeAuthorize))
    ;
    EXPECT_CALL(*mock_request, GetApiVersion())
    .WillOnce(Return("v1000.2"))
    ;

    auto err = handler.ProcessRequest(mock_request.get());
    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kUnsupportedClient));
}




TEST_F(AuthorizerHandlerTests, DataTransferRequestAuthorizeUsesCachedValue) {
    config.authorization_interval_ms = 10000;
    SetReceiverConfig(config, "none");
    MockFirstAuthorization(false);
    EXPECT_CALL(*mock_request, GetOpCode())
    .WillOnce(Return(asapo::kOpcodeTransferData));
    EXPECT_CALL(mock_http_client, Post_t(_, _, _, _, _)).Times(0);
    EXPECT_CALL(*mock_request, SetBeamtimeId(expected_beamtime_id));
    EXPECT_CALL(*mock_request, SetBeamline(expected_beamline));
    EXPECT_CALL(*mock_request, SetDataSource(expected_data_source));
    EXPECT_CALL(*mock_request, SetOnlinePath(expected_beamline_path));
    EXPECT_CALL(*mock_request, SetOfflinePath(expected_core_path));
    EXPECT_CALL(*mock_request, SetSourceType(expected_source_type));
    auto err =  handler.ProcessRequest(mock_request.get());

    ASSERT_THAT(err, Eq(nullptr));
}




}
